@@ -2,7 +2,9 @@
 # -- kongodb --
 #-----------------------------
 
-from . import adapters, jsonquery, lib, database, cursor, exceptions
+from . import adapters, jsonquery, lib, database, cursor, exceptions, types
 
 # alias to Database
 kongodb = database.Database
+
+
