@@ -1,74 +1,16 @@
-#-----------------------------
+# -----------------------------
 # -- kongodb --
 # database module
-#-----------------------------
+# -----------------------------
+
 
-import copy
-import multiprocessing
-from typing import Any, List, Union
 from urllib.parse import urlparse
-from . import adapters, lib, cursor, jsonquery
-
-def parse_row_to_dict(row:dict) -> dict:
-    """
-    Convert a result row to dict, by merging _json with the rest of the columns
-    
-    Args:
-        row: dict
-
-    Returns
-        dict
-    """
-    row = row.copy()
-    _json = lib.json_loads(row.pop("_json")) if "_json" in row else {}
-    return {
-        **row, # ensure columns exists
-        **_json
-    }
-
-def parse_smart_filtering(filters: dict, indexed_columns:list=[]) -> dict:
-    """
-    Smart Filter
-    Breaks down the filters based on type. 
-    - SQL_FILTERS: is more restrictive on what can be queried. 
-                  Will be done at the SQL level
-    - JSON_FILTERS: is more loose. 
-                    It contains items that are not in the sql_filters. 
-                    Will be done at the JSON level
-    
-    Args:
-      filters: dict  - 
-      indexed_columns: list - List of indexed sql columns/or other columns in the table 
-                       this will allow with the smart filtering
-      
-    Returns: 
-      dict:
-        SQL_FILTERS
-        JSON_FILTERS
-    """
-
-    # filter SQL_OPERATORS filters
-    sql_filters = []
-    json_filters = {}
-    for k, v in filters.items():
-        if jsonquery.FILTER_OPERATOR in k:
-            f, op = k.split(jsonquery.FILTER_OPERATOR)
-            if f in indexed_columns and op in jsonquery.SQL_OPERATORS:
-                sql_filters.append((f, jsonquery.SQL_OPERATORS[op], v))
-                continue
-        else:
-            if k in indexed_columns:
-                sql_filters.append((k, jsonquery.SQL_OPERATORS["eq"], v))
-                continue
-        json_filters[k] = v
-              
-    return {
-        "SQL_FILTERS": sql_filters,
-        "JSON_FILTERS": json_filters
-    }
+from typing import Any, List, Union
+from . import adapters, lib, cursor, jsonquery, types, operators
 
 
-DB_DATE_FORMAT = "YYYY-MM-DD HH:mm:ss"
+SQL_DATETIME_FORMAT = "YYYY-MM-DD HH:mm:ss"
+
 
 class Database(object):
     """
@@ -77,7 +19,7 @@ class Database(object):
     Class to create a connection to an adapter and select a container
     """
 
-    def __init__(self, dsn_adapter:Union[str, adapters.BaseAdapter]):
+    def __init__(self, dsn_adapter: Union[str, adapters.BaseAdapter]):
         """
         Connect database 
 
@@ -100,8 +42,7 @@ class Database(object):
             elif _.scheme in ["mysql", "mariadb"]:
                 self.db = adapters.MySQLAdapter(dsn_adapter)
 
-
-    def container(self, name:str, columns:list=None):
+    def container(self, name: str, columns: list = None):
         """
         Select a container
 
@@ -148,13 +89,12 @@ class Item(dict):
     Every row is a document in Kongodb
     """
 
-    def __init__(self, container, row:dict):
-        self.container = container 
+    def __init__(self, container, row: dict):
+        self.container = container
         self._load(row)
 
-    #--- Basic RUD (read|update|delete)
 
-    def get(self, path:str, default:Any=None)->Any:
+    def get(self, path: str, default: Any = None) -> Any:
         """
         GET: Return a property by key/DotNotation
 
@@ -164,13 +104,13 @@ class Item(dict):
         Args:
             path:str - the dotnotation path
             default:Any - default value 
-        
+
         Returns:
             Any
         """
         return lib.dict_get(obj=dict(self), path=path, default=default)
 
-    def set(self, path:str, value:Any):
+    def set(self, path: str, value: Any):
         """
         SET: Set a property by key/DotNotation
 
@@ -181,11 +121,20 @@ class Item(dict):
         Returns:
             Void
         """
-        data = copy.deepcopy(dict(self))
-        lib.dict_set(obj=data, path=path, value=value)
-        self.update(data)
+        self._update({path : value })
 
-    def unset(self, path:str):
+    def len(self, path: str):
+        """
+        Get the length of the items in a str/list/dict
+        Args:
+            path:str - the dotnotation path
+        Returns:
+            data that was removed
+        """
+        v = self.get(path)
+        return len(v) if v else 0
+
+    def unset(self, path: str):
         """ 
         UNSET: Remove a property by key/DotNotation and return the value
 
@@ -195,33 +144,16 @@ class Item(dict):
         Returns:
             Any: the value that was removed
         """
-        data = copy.deepcopy(dict(self))
-        v = lib.dict_pop(obj=data, path=path) if "." in path else data.pop(path)
-        self.update(data)
-        return v
-
-    def update(self, *a, **kw):
+        self._update({ path + ":$unset" % path: True })
+        
+    def update(self, data:dict):
         """
         UPDATE: Update the active Item
 
-        ie:
-            #update(key=value, key2=value2, ...)
-            #update({ "key": value, "key2": value2 })
-        
-        Args:
-            *args
-            **kwargs
-
         Returns:
             Item
         """
-        data = {}
-        if a and isinstance(a[0], dict):
-            data.update(a[0])
-        data.update(kw)
-
-        row = self.container.update(_id=self._id, doc=data, _as_document=False)
-        self._load(row)
+        self._update(data)
 
     def delete(self):
         """
@@ -230,54 +162,38 @@ class Item(dict):
         Returns:
             None
         """
-        self.container.delete(self._id)
-        self._empty_self()
+        self.container.delete_id(self._id)
+        self._clear_self()
 
     # ---
 
-    def _req_int_prop(self, path:str) -> int:
-        v = self.get(path)
-        if v is None:
-            v = 0
-        if not isinstance(v, int):
-            raise TypeError("Invalid data type for '%s'. Must be 'int' " % path)
-        return v
-        
-    def incr(self, path:str, inc=1):
+    def incr(self, path: str, incr=1):
         """
         INCR: increment a value by 1
         Args
             path:str - path
-            inc:1 - value to inc by
+            incr:1 - value to inc by
         Returns:    
             int - the value that was incremented
         """
-        self.set(path=path, value=self._req_int_prop(path) + inc)
-        return self.get(path)
+        op = "%s:$incr" % path
+        oplog = self._update({op: incr })
+        return oplog.get(op)
 
-    def decr(self, path:str, dec=1):
+    def decr(self, path: str, decr=1):
         """
         DECR: decrement a value by 1
         Args
             path:str - path
-            dec:1 - value to dec by
+            decr:1 - value to dec by
         Returns:    
             int - the value that was decremented
-        """   
-        self.set(path=path, value=self._req_int_prop(path) - dec)
-        return self.get(path)
+        """
+        op = "%s:$decr" % path
+        oplog = self._update({op: decr })
+        return oplog.get(op)
 
-    # -- List operations 
-
-    def _req_list_prop(self, path:str) -> list:
-        v = self.get(path)
-        if v is None:
-            return []
-        if not isinstance(v, list):
-            raise TypeError("Invalid data type for '%s'. Must be 'list' " % path)
-        return v
-
-    def ladd(self, path:str, *values:List[Any]):
+    def xadd(self, path: str, values):
         """
         LADD: Add *values if they don't exist yet
 
@@ -287,14 +203,23 @@ class Item(dict):
         Returns:
             list: updated data
         """
-        v = self._req_list_prop(path)
-        for val in values:
-            if val not in v:
-                v.append(val)
-        self.set(path=path, value=v)
-        return self.get(path)
+        op = "%s:$xadd" % path
+        self._update({op: values })
 
-    def lremove(self, path:str, *values:List[Any]):
+    def xadd_multi(self, path: str, *values: List[Any]):
+        """
+        LADD: Add *values if they don't exist yet
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xadd_multi" % path
+        self._update({op: values })
+
+    def xrem(self, path: str, values):
         """
         LREM: Remove items from a list
 
@@ -304,18 +229,25 @@ class Item(dict):
         Returns:
             list: updated data
         """
-        v = self._req_list_prop(path)
-        _removed = False
-        for val in values:
-            if val in v:
-                _removed = True
-                v.remove(val)
-        if _removed:
-            self.set(path=path, value=v)
-            return self.get(path)
-        return v
+        op = "%s:$xrem" % path
+        oplog = self._update({op: values })
+        return oplog.get(op)
 
-    def lpush(self, path:str, *values:List[Any]):
+    def xrem_multi(self, path: str, *values: List[Any]):
+        """
+        LREM: Remove items from a list
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xrem_multi" % path
+        oplog = self._update({op: values })
+        return oplog.get(op)
+
+    def xpush(self, path: str, values:Any):
         """
         LPUSH: push item to the right of list. 
 
@@ -325,12 +257,23 @@ class Item(dict):
         Returns:
             list: updated data
         """
-        v = self._req_list_prop(path)
-        v.extend(values)
-        self.set(path=path, value=v)
-        return self.get(path)
+        op = "%s:$xpush" % path
+        self._update({op: values })
 
-    def lpushl(self, path:str, *values:List[Any]):
+    def xpush_multi(self, path: str, *values: List[Any]):
+        """
+        LPUSH: push item to the right of list. 
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xpush_multi" % path
+        self._update({op: values })
+
+    def lpushl(self, path: str, *values: List[Any]):
         """
         LPUSHL: push item to the left
 
@@ -346,7 +289,7 @@ class Item(dict):
         self.set(path=path, value=v2)
         return self.get(path)
 
-    def lpop(self, path:str): 
+    def lpop(self, path: str):
         """
         Remove value at the end an array/list
         Args:
@@ -355,13 +298,11 @@ class Item(dict):
             data that was removed
 
         """
-        v = self._req_list_prop(path)
-        if len(v):
-            self.set(path=path, value=v[:-1])
-            return v[-1]
-        return None 
+        op = "%s:$xpop" % path
+        oplog = self._update({op: True })
+        return oplog.get(op)
 
-    def lpopl(self, path:str): 
+    def lpopl(self, path: str):
         """
         Remove value at the beginning an array/list
         Args:
@@ -369,22 +310,10 @@ class Item(dict):
         Returns:
             data that was removed        
         """
-        v = self._req_list_prop(path)
-        if len(v):
-            self.set(path=path, value=v[1:])  
-            return v[0]
-        return None
+        op = "%s:$xpopl" % path
+        oplog = self._update({op: True })
+        return oplog.get(op)
 
-    def len(self, path:str):
-        """
-        Get the length of the items in a list/object/dict
-        Args:
-            path:str - the dotnotation path
-        Returns:
-            data that was removed
-        """
-        v = self.get(path)
-        return len(v) if v else 0
 
     # --- misc
     def save(self):
@@ -397,19 +326,29 @@ class Item(dict):
         data = dict(self)
         self.update(data)
 
-    def _load(self, row:dict):
+    def _update(self, ops:dict):
+        """
+        Return oplog
+        """
+        data = dict(self)
+        doc, oplog = operators.update_operations(data, ops)
+        row = self.container.update_id(_id=self._id, doc=doc, _as_document=False)
+        self._load(row)
+        return oplog
+
+    def _load(self, row: dict):
         """
         load the content into the document
-        
+
         Args:
             row: dict
         """
-        self._empty_self()
-        row = parse_row_to_dict(row)
+        self._clear_self()
+        row = _parse_row(row)
         self._id = row.get("_id")
         super().__init__(row)
 
-    def _empty_self(self):
+    def _clear_self(self):
         """ clearout all properties """
         for _ in list(self.keys()):
             if _ in self:
@@ -426,7 +365,7 @@ class Container(object):
     _columns = []
     _indexes = []
 
-    def __init__(self, conn:adapters.BaseAdapter, name):
+    def __init__(self, conn: adapters.BaseAdapter, name):
         self.name = name
         self.db = conn
         self.db.create_container(self.name)
@@ -468,7 +407,7 @@ class Container(object):
 
     # ---- methods ----
 
-    def get(self, _id:str) -> Item:
+    def get(self, _id: str) -> Item:
         """
         Get a document by _id
         Alias to find_one(_id)
@@ -491,6 +430,7 @@ class Container(object):
           Item
 
         Examples:
+            #.find_one()
             #.find_one('IDfyeiquyteqtyqiuyt')
             #.find_one(_id="ID....")
             #.find_one(key1=value1, key2=value2, ...)
@@ -503,9 +443,8 @@ class Container(object):
             _as_document = kw.pop("_as_document")
 
         # expecting the first arg to be _id
-        filters =  {"_id": a[0]} if a else kw
-        if not filters:
-            raise Exception("Invalid Container.get args")
+        filters = {"_id": a[0]} if a else kw
+
         r = self.find(filters=filters, limit=1, _as_document=_as_document)
         return r[0] if len(r) else None
 
@@ -518,17 +457,17 @@ class Container(object):
         """
         return list(self.find(*a, **kw))
 
-    def find(self, filters: dict = {}, sort: list = [], limit: int = 10, skip: int = 0, _as_document=True) -> cursor.Cursor:
+    def find(self, filters: dict = {}, sorts:dict = {}, limit: int = 10, skip: int = 0, _as_document=True) -> cursor.Cursor:
         """
         To fetch data from the container
 
         Smart Query
           Allow to use primary indexes from sqlite 
           then do the xtra from parsing the documents
-          
+
         Args:
           filters:dict - 
-          sort:list - [(column, order[-1|1])]
+          sort:dict - {key: "DESC", key2: "ASC"}
           limit:int - 
           skit:int - 
 
@@ -541,9 +480,10 @@ class Container(object):
         # If there is more search properties, take it to the json
         xparams = []
         xquery = []
-
-        smart_filters = parse_smart_filtering(filters, indexed_columns=self.columns)
         
+        smart_filters = _parse_smart_filtering(
+            filters, indexed_columns=self.columns)
+
         # Build the SQL query
         query = "SELECT * FROM %s " % self.name
 
@@ -558,7 +498,6 @@ class Container(object):
                     xparams.append(f[2])
         if xquery and xparams:
             query += " WHERE %s " % " AND ".join(xquery)
-            
 
         # Perform JSON search, as we have JSON_FILTERS
         # Full table scan, relative to WHERE clause
@@ -567,7 +506,7 @@ class Container(object):
         if smart_filters["JSON_FILTERS"]:
             for chunked in self.db.fetchmany(query, xparams, chunk):
                 if chunked:
-                    rows = [parse_row_to_dict(row) for row in chunked]
+                    rows = [_parse_row(row) for row in chunked]
                     for r in jsonquery.execute(rows, smart_filters["JSON_FILTERS"]):
                         data.append(r)
                 else:
@@ -575,36 +514,37 @@ class Container(object):
             if data:
                 if _as_document:
                     data = [Item(self, d) for d in data]
-            return cursor.Cursor(data, sort=sort, limit=limit, skip=skip)
+            sorts = _parse_sort_dict(sorts, False)
+            return cursor.Cursor(data, sort=sorts, limit=limit, skip=skip)
 
         # Skip JSON SEARCH, use only SQL.
         # No need to look into the JSON. The DB is enough
         else:
             # order by
-            if sort:
+            sorts = _parse_sort_dict(sorts, True)
+            if sorts:
                 query += " ORDER BY "
-                for _ in sort:
-                    query += " %s %s" % (_[0], "DESC" if _[0] == -1 else "ASC")
-           
+                query += " ".join([" %s %s " % (_[0], _[1]) for _ in sorts])
+
             # limit/skip
             if limit or skip:
                 query += " LIMIT ?, ?"
                 xparams.append(skip or 0)
                 xparams.append(limit or 10)
 
-            res = self.db.fetchall(query, xparams) 
-            if _as_document:           
+            res = self.db.fetchall(query, xparams)
+            if _as_document:
                 data = [Item(self, row) for row in res]
             else:
                 data = list(res)
             return cursor.Cursor(data)
 
-    def add(self, doc: dict, _as_document:bool=True) -> Item:
+    def add(self, doc: dict, _as_document: bool = True) -> Item:
         """
         Add a new document in container
 
         use Smart Insert, by checking if a value in the doc in is a column.
-        
+
         Args:
           doc:dict - Data to be inserted
 
@@ -616,8 +556,8 @@ class Container(object):
 
         _id = lib.gen_id()
         ts = lib.get_timestamp()
-        f_ts = ts.format(DB_DATE_FORMAT)
-        doc.upate({
+        f_ts = ts.format(SQL_DATETIME_FORMAT)
+        doc.update({
             "_id": _id,
             "_created_at": ts,
             "_modified_at": ts,
@@ -626,20 +566,24 @@ class Container(object):
         xcolumns = ["_id", "_json", "_created_at", "_modified_at"]
         xparams = [_id, lib.json_dumps(doc), f_ts, f_ts]
         q = "INSERT INTO %s " % self.name
-        
+
         # indexed data
-        # some data can't be overwritten 
+        # some data can't be overwritten
         for col in self.columns:
             if col in doc and col not in xcolumns:
                 xcolumns.append(col)
                 xparams.append(doc[col])
 
-        q += " ( %s ) VALUES ( %s ) " % (",".join(xcolumns), ",".join(["?" for _ in xparams]))
-        
+        q += " ( %s ) VALUES ( %s ) " % (",".join(xcolumns),
+                                         ",".join(["?" for _ in xparams]))
+
         self.db.execute(q, xparams)
         return self.find_one(_id=_id, _as_document=_as_document)
 
-    def update(self, _id: str, doc: dict = {}, replace: bool = False, _as_document=True) -> Item:
+    def update(self, filters): raise NotImplementedError()
+    def delete(self, filters): raise NotImplementedError()
+
+    def update_id(self, _id: str, doc: dict = {}, replace: bool = False, _as_document=True) -> Item:
         """
         To update a document
 
@@ -655,8 +599,9 @@ class Container(object):
         rdoc = self.find_one(_id=_id, _as_document=False)
         if rdoc:
             ts = lib.get_timestamp()
-            ts_f = ts.format(DB_DATE_FORMAT)
-            _doc = doc if replace else lib.dict_merge(lib.json_loads(rdoc["_json"]), doc)
+            ts_f = ts.format(SQL_DATETIME_FORMAT)
+            _doc = doc if replace else lib.dict_merge(
+                lib.json_loads(rdoc["_json"]), doc)
             _doc["_modified_at"] = ts
             _restricted_columns = self.DEFAULT_COLUMNS[:]
 
@@ -664,9 +609,9 @@ class Container(object):
             xparams = [lib.json_dumps(_doc), ts_f]
 
             q = "UPDATE %s SET " % self.name
-            
+
             # indexed data
-            # some data can't be overriden 
+            # some data can't be overriden
             for col in self.columns:
                 if col in _doc and col not in _restricted_columns:
                     xcolumns.append(col)
@@ -678,10 +623,10 @@ class Container(object):
             return self.find_one(_id=_id, _as_document=_as_document)
         return None
 
-    def delete(self, _id: str) -> bool:
+    def delete_id(self, _id: str) -> bool:
         """
         To delete an entry by _id
-        
+
         Args:
             _id:str - entry id
 
@@ -691,45 +636,23 @@ class Container(object):
         self.db.execute("DELETE FROM %s WHERE _id=?" % (self.name), (_id, ))
         return True
 
-    def add_columns(self, columns:List[str], enforce_index=False):
+    def add_columns(self, columns: list, enforce_index=False):
         """
         To add columns. With options to add indexes
 
         Args:
-            columns: 
-                shortform: "NAME:TYPE@INDEX"
-                longform: 'NAME:TYPE=extra@INDEX',
-                List[str] -> "COLUMN:TYPE@INDEX"
-                    [
-                        "column", # column only. Type is in
-                        "column:type", # column and type
-                        "column:type@index", # column, type and index
-                        "column:type@unique" # column type and unique index
-                        "column@unique" # column and unique index. Type is inferred
-                    ]
-            
-            enforce_index:
-                bool - To make all prop an index.
-        Returns:
-            None
+            columns: List[str|types.BaseType]
         """
         cols_stmt = []
-        for idx in columns:
-            if isinstance(idx, str):
-                _type = "TEXT"
-                indx = False
-                col = idx
-                if "@" in col:
-                    col, indx =  col.split("@")
-                    indx = "UNIQUE" if indx.upper() == "UNIQUE" else True
-                if ":" in col:
-                    col, _type = col.split(":")
-                if enforce_index and indx != "UNIQUE":
-                    indx = True
-                cols_stmt.append((col, _type or "TEXT", indx))
+        for stmt in columns:
+            if isinstance(stmt, types.BaseType):
+                cols_stmt.append(stmt)
+            elif isinstance(stmt, str):
+                cols_stmt.append(types.stmt_to_fieldtype(stmt))
+
         self.db.add_columns(table=self.name, cols_stmt=cols_stmt)
-        
-    def add_indexes(self, columns:List[str]):
+
+    def add_indexes(self, columns: List[str]):
         """
         To indexed columns
 
@@ -739,7 +662,81 @@ class Container(object):
         """
         self.add_columns(columns=columns, enforce_index=True)
 
-    #TODO: 
+    # TODO:
     # def insert_many(self, data:List[dict])
     # def update_many(self, data:dict, filters:dict)
-    # def delete_many()
\ No newline at end of file
+    # def delete_many()
+
+
+
+def _parse_smart_filtering(filters: dict, indexed_columns: list = []) -> dict:
+    """
+    Smart Filter
+    Breaks down the filters based on type. 
+    - SQL_FILTERS: is more restrictive on what can be queried. 
+                  Will be done at the SQL level
+    - JSON_FILTERS: is more loose. 
+                    It contains items that are not in the sql_filters. 
+                    Will be done at the JSON level
+
+    Args:
+      filters: dict  - 
+      indexed_columns: list - List of indexed sql columns/or other columns in the table 
+                       this will allow with the smart filtering
+
+    Returns: 
+      dict:
+        SQL_FILTERS
+        JSON_FILTERS
+    """
+
+    # filter SQL_OPERATORS filters
+    sql_filters = []
+    json_filters = {}
+    for k, v in filters.items():
+        if jsonquery.FILTER_OPERATOR in k:
+            f, op = k.split(jsonquery.FILTER_OPERATOR)
+            if f in indexed_columns and op in jsonquery.SQL_OPERATORS:
+                sql_filters.append((f, jsonquery.SQL_OPERATORS[op], v))
+                continue
+        else:
+            if k in indexed_columns:
+                sql_filters.append((k, jsonquery.SQL_OPERATORS["eq"], v))
+                continue
+        json_filters[k] = v
+
+    return {
+        "SQL_FILTERS": sql_filters,
+        "JSON_FILTERS": json_filters
+    }
+
+
+def _parse_row(row: dict) -> dict:
+    """
+    Convert a result row to dict, by merging _json with the rest of the columns
+
+    Args:
+        row: dict
+
+    Returns
+        dict
+    """
+    row = row.copy()
+    _json = lib.json_loads(row.pop("_json")) if "_json" in row else {}
+    return {
+        **row,  # ensure columns exists
+        **_json
+    }
+
+def _ascdesc(v, as_str=True):
+    if as_str:
+        if isinstance(v, int):
+            return "DESC" if v == -1 else "ASC" 
+    else:
+        if isinstance(v, str):
+            return -1 if v.upper() == "DESC" else 1
+    return v
+
+def _parse_sort_dict(sorts:dict, as_str=True):
+    return [(k, _ascdesc(v, as_str)) for k,v in sorts.items()]
+
